/* cpu3 simulator version 1.0

simulate cpu3 instruction set


implement Big endian
jsr 1000  print ac
jsr 1001  stop the simulator
id        return 99
not implement :  interrupt, ina, out

input takes S-format file

Prabhas Chongstitvatana  8th September 1999
*/

#include <stdio.h>

#include "opcode.h"
#include "hex.c"        /* for S-format reader writer */

#define  MXMEM	1000

#define  IM	1	/* addressing mode */
#define  DI	2
#define	 IX	3

#define  PRINT  1000
#define  STOP	1001

typedef struct { char name[5]; char gr; } optype;
optype	optable[] =
	{"add",1, "adc",1, "com",0, "and",1, "ora",1, "xor",1,
	 "cla",0, "lda",1, "sta",2, "psh",0, "pop",0, "nop",0,
	 "ina",3, "out",3, "jz",2,  "jc",2,  "jn",2,  "jmp",2,
	 "jsr",2, "rts",0, "rti",0, "ldx",4, "stx",2, "inx",0,
	 "dex",0, "tsx",0, "rol",0, "clc",0, "stc",0, "cli",0,
	 "sti",0, "ror",0, "id",0   };

char M[MXMEM];
char Cf, Zf, Nf, If;	/* flags : carry, zero, neg, int */
int  ac, pc;
long sp, rx;

int op, trace;

error( char *s )
{
  printf("error %s at pc = %d op = %d\n", s,pc,op);
  exit(0);
}

int get2( int ads)
{
  int v;
  unsigned char c1, c2;
  c1 = M[ads];
  c2 = M[ads+1];
  v = 256*c1 + c2;
  return v;
}

put2( int ads, int value)
{
  M[ads] = (value >> 8) & 255;
  M[ads+1] = value & 255;
}

flagZN()
{
  Zf = ( ac == 0 ) ? 1 : 0;
  Nf = ( ac & 0x080 ) ? 1 : 0;
}

allflag()
{
  Cf = ( ac > 255 ) ? 1 : 0;
  if( Cf ) ac &= 255;
  flagZN();
}

xflagZN()
{
  Zf = ( rx == 0 ) ? 1 : 0;
  Nf = ( rx & 0x0800) ? 1 : 0;
}

xallflag()
{
  Cf = ( rx > 65535 ) ? 1 : 0;
  if( Cf ) rx &= 0xffff;
  xflagZN();
}

int inPort( int portNo ) {}
outPort( int portNo, int value) {}

push( char x )
{
  M[sp] = x; sp--;
}

char pop()
{
  sp++; return M[sp];
}

push2( int x )  /* push lo then push hi */
{
  push( x & 255 );
  push( (x << 8) & 255);
}

int pop2()  /* pop hi then pop lo */
{
  int hi,lo;
  hi = pop();
  lo = pop();
  return (int)( (hi << 8) | ( lo & 255) );
}

/* the instructions are groups according to how
the operand is fetched.
group 0 no operand
group 1 fetch value 8 bits : ADD,ADC,AND,ORA,XOR,LDA
group 2 fetch effective address 16 bits : STA,JZ,JC,JN,JMP,JSR,STX
group 3 fetch eff ads 8 bits (in out) : INA,OUT
group 4 fetch value 16 bits : LDX
*/

int fetchOperand( int op, int mode )
{
  int v, ea, ofs;

  switch( optable[op].gr ) {
  case 0 : return 0;
  case 1 :
    switch(mode) {
    case IM : v = M[pc]; pc++; break;
    case DI : ea = get2(pc); v = M[ea]; pc += 2; break;
    case IX : ofs = M[pc]; pc++;
	      v = M[ rx + ofs ]; break;
    }
    return v & 255;
  case 2 :
    switch( mode ) {
    case IM : error("illegal addressing mode"); break;
    case DI : ea = get2(pc); pc += 2; break;
    case IX : ea = rx + M[pc]; pc++ ; break;
    }
    return ea;
  case 3 :
    switch( mode ) {
    case IM : error("illegal addressing mode"); break;
    case DI : ea = M[pc]; pc++; break;
    case IX : ea = (rx + M[pc]); pc++; break;
    }
    return ea & 255;
  case 4 :
    switch( mode ) {
    case IM : v = get2(pc); pc += 2; break;
    case DI : ea = get2(pc); v = get2(ea); pc+= 2; break;
    case IX : ea = rx + M[pc]; v = get2(ea); pc++; break;
    }
    return v;
  }
}

trap(int v)  /* special trap 1000 print ac, 1001 stop */
{
  switch( v ) {
  case PRINT : printf("ac %d\n",ac); break;
  case STOP  : exit(0);
  default    : push2(pc); pc = v;
  }
}

run(int limit)  /* limit the number of instructions
		   to prevent the infinite loop */
{
  int ir, mode, v, c2;

  while(limit--) {
    ir = M[pc++];
    op = (ir>>2) & 63;
    mode = ir & 3;
    v = fetchOperand(op,mode);
    switch( op ) {
    case ADD:  ac += v; allflag(); break;
    case ADC:  ac += (v + Cf); allflag(); break;
    case COM:  ac = ~ac; ac &= 255; flagZN(); break;
    case AND:  ac &= v; flagZN(); break;
    case ORA:  ac |= v; flagZN(); break;
    case XOR:  ac ^= v; flagZN(); break;
    case CLA:  ac = 0; flagZN(); break;
    case LDA:  ac = v; flagZN(); break;
    case STA:  M[v] = ac; break;
    case PSH:  push(ac); break;
    case POP:  ac = pop(); flagZN(); break;
    case NOP:  break;
    case INA:  ac = inPort(v); flagZN(); break;
    case OUT:  outPort(v,ac); break;
    case JZ:   if( Zf ) pc = v; break;
    case JC:   if( Cf ) pc = v; break;
    case JN:   if( Nf ) pc = v; break;
    case JMP:  pc = v; break;
    case JSR:  trap(v); break;
    case RTS:  pc = pop2(); break;
    case RTI:  break;
    case LDX:  rx = v &0x0ffff; xflagZN(); break;
    case STX:  put2(v,rx); break;
    case INX:  rx++; xallflag(); break;
    case DEX:  rx--; xallflag(); break;
    case TSX:  rx = pop2() & 0x0ffff; xallflag(); break;
    case ROL:  ac = (ac << 1) | Cf;
	       Cf = ( ac & 0x080 ) ? 1 : 0;
	       ac &= 255;
	       flagZN();
	       break;
    case CLC:  Cf = 0; break;
    case STC:  Cf = 1; break;
    case CLI:  If = 0; break;
    case STI:  If = 1; break;
    case ROR:  c2 = ac & 1;
	       ac = ac >> 1;
	       if( Cf )  ac |= 0x080;
	       ac &= 255;
	       Cf = c2;
	       flagZN();
	       break;
    case ID:   ac = 99; break;
    default:   error("undefine instruction");
    }
    if( trace ) show();
  }
}

show()
{
/*  printf("pc %4X %3s A %2X X %4lX SP %4lX INCZ %d%d%d%d\n",
	  pc,optable[op].name,ac,rx,sp,If,Nf,Cf,Zf);
*/
  printf("pc %4d %3s A %3d X %5ld SP %5ld INCZ %d%d%d%d\n",
	  pc,optable[op].name,ac,rx,sp,If,Nf,Cf,Zf);

}

main()
{
  int a;
  trace = 1;
  a = Sreader(M);
  printf("last address loaded %d\n",a);

  pc = 0; Cf = 0; Zf = 0; Nf = 0; If = 0;
  ac = 0; sp = MXMEM-1; rx = 0;
  run(500);
}

